/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.decompose;

import java.util.ArrayList;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IHierarchicElementContainer;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.kernel.extension.IElementCompositor;
import org.fortiss.tooling.kernel.service.IConnectionCompositorService;
import org.fortiss.tooling.kernel.service.IElementCompositorService;

/**
 * Base implementation of a compositor for {@link IHierarchicElementContainer}s
 * focusing on the remove mechanism.
 * 
 * @author doebber
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating GREEN Hash: 88EAEE8F01C0BDB67D8205D322A68EAF
 */
public abstract class HierarchicElementCompositorBase<HE extends IHierarchicElementContainer>
		implements IElementCompositor<HE> {

	/**
	 * Decomposition that takes care of the whole model subtree. Subclasses
	 * implement {@link #decompose*****()} instead.
	 */
	@SuppressWarnings("javadoc")
	@Override
	public final boolean decompose(EObject contained) {
		return iterateDecompose(contained);
	}

	/**
	 * Base implementation for element removal using EcoreUtil.delete
	 * (non-recursively). Subclasses may override.
	 */
	protected boolean decomposeSpecific(EObject contained) {
		EcoreUtil.delete(contained);
		return true;
	}

	/** recursive traversion of all model elements in subtree */
	private boolean iterateDecompose(EObject contained) {
		if(contained instanceof IModelElement) {
			if(!iterateDecomposeSpecifications((IModelElement)contained)) {
				return false;
			}
			// TODO (CD): https://af3.fortiss.org/issues/389
			// handle reference removal
		}
		if(contained instanceof IHierarchicElementContainer) {
			if(!iterateDecomposeSubelements((IHierarchicElementContainer)contained)) {
				return false;
			}
		}
		if(contained instanceof IHierarchicElement) {
			IHierarchicElement he = (IHierarchicElement)contained;

			if(!iterateDecomposeConnections(he.getConnectionsList())) {
				return false;
			}

			if(!iterateDecomposeConnectors((IHierarchicElement)contained)) {
				return false;
			}

		}
		if(contained instanceof IConnector) {
			IConnector connector = (IConnector)contained;
			if(!iterateDecomposeConnections(connector.getIncomingList())) {
				return false;
			}
			if(!iterateDecomposeConnections(connector.getOutgoingList())) {
				return false;
			}
		}
		return decomposeSpecific(contained);
	}

	/**
	 * Iterates over specifications of given {@link IModelElement} and performs
	 * decomposition.
	 */
	private boolean iterateDecomposeSpecifications(IModelElement me) {
		ArrayList<IModelElementSpecification> list =
				new ArrayList<IModelElementSpecification>(me.getSpecificationsList());
		for(IModelElementSpecification spec : list) {
			if(!decomposeSpecification(spec)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over sub elements of given {@link IHierarchicElementContainer} and performs
	 * decomposition.
	 */
	private boolean iterateDecomposeSubelements(IHierarchicElementContainer contained) {
		ArrayList<IHierarchicElement> list =
				new ArrayList<IHierarchicElement>(contained.getContainedElementsList());
		for(IHierarchicElement subelement : list) {
			if(!decomposeSubelement(subelement)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over list of connections and runs disconnection or decomposition
	 * of sub elements respectively.
	 */
	private boolean iterateDecomposeConnections(EList<IConnection> connectionsList) {
		ArrayList<IConnection> list = new ArrayList<IConnection>(connectionsList);
		for(IConnection conn : list) {
			if(!decomposeConnection(conn)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Iterates over connectors of given {@link IHierarchicElement} and runs
	 * decomposition.
	 */
	private boolean iterateDecomposeConnectors(IHierarchicElement contained) {
		ArrayList<IConnector> list = new ArrayList<IConnector>(contained.getConnectorsList());
		for(IConnector c : list) {
			if(!decomposeConnector(c)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Base implementation uses element compositor service's decomposition by
	 * default and returns <code>true</code>. Subclasses may override.
	 */
	protected boolean decomposeSpecification(IModelElementSpecification element) {
		if(IElementCompositorService.INSTANCE.canDecompose(element)) {
			return IElementCompositorService.INSTANCE.decompose(element);
		}
		// no delete here, since normally specs are destroyed with the element
		return true;
	}

	/** Base implementation yet empty. */
	protected boolean decomposeReferences(@SuppressWarnings("unused") EObject element) {
		// TODO (FH): https://af3.fortiss.org/issues/389
		// implement reference removal
		return true;
	}

	/**
	 * Base implementation returns compositors' decomposition by default and
	 * invokes EcoreUtil.delete() otherwise Subclasses may override.
	 */
	protected boolean decomposeSubelement(IHierarchicElement element) {
		if(IElementCompositorService.INSTANCE.canDecompose(element)) {
			return IElementCompositorService.INSTANCE.decompose(element);
		}
		EcoreUtil.delete(element);
		return true;
	}

	/**
	 * Base implementation returns compositors' decomposition by default and
	 * invokes EcoreUtil.delete() otherwise Subclasses may override.
	 */
	protected boolean decomposeConnector(IConnector element) {
		ArrayList<IConnection> list = new ArrayList<IConnection>(element.getIncomingList());
		list.addAll(element.getOutgoingList());
		for(IConnection conn : list) {
			if(!IConnectionCompositorService.INSTANCE.disconnect(conn)) {
				return false;
			}
		}
		EcoreUtil.delete(element);
		return true;
	}

	/**
	 * Base implementation returns compositors' decomposition by default and
	 * invokes EcoreUtil.delete() otherwise Subclasses may override.
	 */
	protected boolean decomposeConnection(IConnection conn) {
		return IConnectionCompositorService.INSTANCE.disconnect(conn);
	}

}
